@@ -1,24 +1,7 @@
 /*
  * Copyright (C) 2014 - 2016, Stephan Mueller <smueller@chronox.de>
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, and the entire permission notice in its entirety,
- *    including the disclaimer of warranties.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior
- *    written permission.
- *
- * ALTERNATIVELY, this product may be distributed under the terms of
- * the GNU General Public License, in which case the provisions of the GPL2
- * are required INSTEAD OF the above restrictions.  (This clause is
- * necessary due to a potential bad interaction between the GPL and
- * the restrictions contained in a BSD-style copyright.)
+ * License: see COPYING file in root directory
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
@@ -88,10 +71,14 @@ struct kcapi_handle;
  * This function provides the initialization of a symmetric cipher handle and
  * establishes the connection to the kernel.
  *
- * @return 0 upon success; ENOENT - algorithm not available;
- *	    -EOPNOTSUPP - AF_ALG family not available;
- *	    -EINVAL - accept syscall failed
- *	    -ENOMEM - cipher handle cannot be allocated
+ * On success, a pointer to kcapi_handle object is returned in *handle.
+ * Function kcapi_cipher_destroy should be called afterwards to free resources.
+ *
+ * @return 0 upon success;
+ *	   -ENOENT - algorithm not available;
+ *	   -EOPNOTSUPP - AF_ALG family not available;
+ *	   -EINVAL - accept syscall failed
+ *	   -ENOMEM - cipher handle cannot be allocated
  */
 int kcapi_cipher_init(struct kcapi_handle **handle, const char *ciphername,
 		      uint32_t flags);
@@ -117,7 +104,7 @@ void kcapi_cipher_destroy(struct kcapi_handle *handle);
  * as it is now maintained by the kernel.
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_cipher_setkey(struct kcapi_handle *handle,
 			const uint8_t *key, uint32_t keylen);
@@ -146,7 +133,7 @@ int kcapi_cipher_setkey(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes encrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
@@ -157,8 +144,9 @@ int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * kcapi_cipher_encrypt_aio() - encrypt data (asynchronous one shot)
  *
  * @handle: [in] cipher handle
- * @iov: [in/out] head of scatter-gather list array holding the plaintext
- *	which will be replaced with ciphertext
+ * @iniov: [in] head of scatter-gather list array holding the plaintext
+ * @outiov: [out] head of scatter-gather list of the destination buffers filled
+ *	with ciphertext
  * @iovlen: [in] number of scatter-gather list entries
  * @iv: [in] IV to be used for cipher operation
  * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
@@ -175,9 +163,10 @@ int32_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes encrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
-int32_t kcapi_cipher_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+int32_t kcapi_cipher_encrypt_aio(struct kcapi_handle *handle,
+				 struct iovec *iniov, struct iovec *outiov,
 				 uint32_t iovlen, const uint8_t *iv,
 				 int access);
 
@@ -205,7 +194,7 @@ int32_t kcapi_cipher_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes decrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 			     const uint8_t *in, uint32_t inlen,
@@ -216,8 +205,9 @@ int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
  * kcapi_cipher_decrypt_aio() - decrypt data (asynchronous one shot)
  *
  * @handle: [in] cipher handle
- * @iov: [in/out] head of scatter-gather list array holding the ciphertext
- *	which will be replaced with plaintext
+ * @iniov: [in] head of scatter-gather list array holding the ciphertext
+ * @outiov: [out] head of scatter-gather list with the destination buffers for
+ *	the plaintext
  * @iovlen: [in] number of scatter-gather list entries
  * @iv: [in] IV to be used for cipher operation
  * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
@@ -234,9 +224,10 @@ int32_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes decrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
-int32_t kcapi_cipher_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
+int32_t kcapi_cipher_decrypt_aio(struct kcapi_handle *handle,
+				 struct iovec *iniov, struct iovec *outiov,
 				 uint32_t iovlen, const uint8_t *iv,
 				 int access);
 
@@ -267,7 +258,7 @@ int32_t kcapi_cipher_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
 				     const uint8_t *iv,
@@ -299,7 +290,7 @@ int32_t kcapi_cipher_stream_init_enc(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
 				     const uint8_t *iv,
@@ -335,7 +326,7 @@ int32_t kcapi_cipher_stream_init_dec(struct kcapi_handle *handle,
  * will be put to sleep until another thread invokes kcapi_cipher_stream_op().
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
 				   struct iovec *iov, uint32_t iovlen);
@@ -360,7 +351,7 @@ int32_t kcapi_cipher_stream_update(struct kcapi_handle *handle,
  * into the buffer.
  *
  * @return number of bytes obtained from the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
 			       struct iovec *iov, uint32_t iovlen);
@@ -371,7 +362,7 @@ int32_t kcapi_cipher_stream_op(struct kcapi_handle *handle,
  * @handle: [in] cipher handle
  *
  * @return > 0 specifying the IV size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_cipher_ivsize(struct kcapi_handle *handle);
 
@@ -381,7 +372,7 @@ uint32_t kcapi_cipher_ivsize(struct kcapi_handle *handle);
  * @handle: [in] cipher handle
  *
  * @return > 0 specifying the block size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle);
 
@@ -404,11 +395,14 @@ uint32_t kcapi_cipher_blocksize(struct kcapi_handle *handle);
  * This function initializes an AEAD cipher handle and establishes the
  * connection to the kernel.
  *
+ * On success, a pointer to kcapi_handle object is returned in *handle.
+ * Function kcapi_aead_destroy should be called afterwards to free resources.
+ *
  * @return 0 upon success;
- *	    -ENOENT - algorithm not available;
- *	    -EOPNOTSUPP - AF_ALG family not available;
- *	    -EINVAL - accept syscall failed
- *	    -ENOMEM - cipher handle cannot be allocated
+ *	   -ENOENT - algorithm not available;
+ *	   -EOPNOTSUPP - AF_ALG family not available;
+ *	   -EINVAL - accept syscall failed
+ *	   -ENOMEM - cipher handle cannot be allocated
  */
 int kcapi_aead_init(struct kcapi_handle **handle, const char *ciphername,
 		    uint32_t flags);
@@ -434,7 +428,7 @@ void kcapi_aead_destroy(struct kcapi_handle *handle);
  * as it is now maintained by the kernel.
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_aead_setkey(struct kcapi_handle *handle,
 		      const uint8_t *key, uint32_t keylen);
@@ -449,7 +443,7 @@ int kcapi_aead_setkey(struct kcapi_handle *handle,
  * created during encryption operation with the size provided with this call.
  *
  * @return 0 upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_aead_settaglen(struct kcapi_handle *handle, uint32_t taglen);
 
@@ -495,15 +489,15 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, uint32_t assoclen);
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * After invoking this function the caller should use
- * kcapi_aead_getdata() to obtain the resulting ciphertext and authentication
- * tag references.
+ * kcapi_aead_getdata_output() to obtain the resulting ciphertext and
+ * authentication tag references.
  *
  * IMPORTANT NOTE: The kernel will only process
  * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
  * be handled by the kernel.
  *
  * @return number of bytes encrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
@@ -515,7 +509,8 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
  * kcapi_aead_encrypt_aio() - asynchronously encrypt AEAD data (one shot)
  *
  * @handle: [in] cipher handle
- * @iov: [in/out] array of scatter-gather list with input / output buffers
+ * @iniov: [in] array of scatter-gather list with input buffers
+ * @outiov: [out] array of scatter-gather list with output buffers
  * @iovlen: [in] number of IOVECs in array
  * @iv: [in] IV to be used for cipher operation
  * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
@@ -536,18 +531,19 @@ int32_t kcapi_aead_encrypt(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * After invoking this function the caller should use
- * kcapi_aead_getdata() to obtain the resulting ciphertext and authentication
- * tag references.
+ * kcapi_aead_getdata_output() to obtain the resulting ciphertext and
+ * authentication tag references.
  *
  * IMPORTANT NOTE: The kernel will only process
  * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
  * be handled by the kernel.
  *
  * @return number of bytes encrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
-int32_t kcapi_aead_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
-			       uint32_t iovlen, const uint8_t *iv, int access);
+int32_t kcapi_aead_encrypt_aio(struct kcapi_handle *handle, struct iovec *iniov,
+			       struct iovec *outiov, uint32_t iovlen,
+			       const uint8_t *iv, int access);
 
 /**
  * kcapi_aead_getdata() - get the resulting data from encryption
@@ -568,6 +564,7 @@ int32_t kcapi_aead_encrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
  * @taglen: [out] length of tag; when tag was set to NULL, no information
  *	is returned
  *
+ * DEPRECATED since 0.13.0
  *
  * This function is a service function to the consumer to locate the right
  * ciphertext buffer offset holding the authentication tag. In addition, it
@@ -580,6 +577,68 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
 			uint8_t **data, uint32_t *datalen,
 			uint8_t **tag, uint32_t *taglen);
 
+/**
+ * kcapi_aead_getdata_input() - get the pointers into input buffer
+ *
+ * @handle: [in] cipher handle
+ * @encdata: [in] data buffer returned by the encryption operation
+ * @encdatalen: [in] size of the encryption data buffer
+ * @enc: [in] does output buffer hold encryption or decryption result?
+ * @aad: [out] AD buffer pointer;  when set to NULL, no data pointer is
+ *	returned
+ * @aadlen: [out] length of AD; when aad was set to NULL, no information is
+ *	returned
+ * @data: [out] pointer to output buffer from AEAD encryption operation
+ *	when set to NULL, no data pointer is returned
+ * @datalen: [out] length of data buffer; when data was set to NULL, no
+ *	information is returned
+ * @tag: [out] tag buffer pointer;  when set to NULL, no data pointer is
+ *	returned
+ * @taglen: [out] length of tag; when tag was set to NULL, no information
+ *	is returned
+ *
+ * This function is a service function to the consumer to locate the right
+ * ciphertext buffer offset holding the authentication tag. In addition, it
+ * provides the consumer with the length of the tag and the length of the
+ * ciphertext.
+ */
+void kcapi_aead_getdata_input(struct kcapi_handle *handle,
+			      uint8_t *encdata, uint32_t encdatalen, int enc,
+			      uint8_t **aad, uint32_t *aadlen,
+			      uint8_t **data, uint32_t *datalen,
+			      uint8_t **tag, uint32_t *taglen);
+
+/**
+ * kcapi_aead_getdata_output() - get the pointers into output buffer
+ *
+ * @handle: [in] cipher handle
+ * @encdata: [in] data buffer returned by the encryption operation
+ * @encdatalen: [in] size of the encryption data buffer
+ * @enc: [in] does output buffer hold encryption or decryption result?
+ * @aad: [out] AD buffer pointer;  when set to NULL, no data pointer is
+ *	returned; returned pointer may also be NULL
+ * @aadlen: [out] length of AD; when aad was set to NULL, no information is
+ *	returned
+ * @data: [out] pointer to output buffer from AEAD encryption operation
+ *	when set to NULL, no data pointer is returned
+ * @datalen: [out] length of data buffer; when data was set to NULL, no
+ *	information is returned
+ * @tag: [out] tag buffer pointer;  when set to NULL, no data pointer is
+ *	returned; returned pointer may also be NULL
+ * @taglen: [out] length of tag; when tag was set to NULL, no information
+ *	is returned
+ *
+ * This function is a service function to the consumer to locate the right
+ * ciphertext buffer offset holding the authentication tag. In addition, it
+ * provides the consumer with the length of the tag and the length of the
+ * ciphertext.
+ */
+void kcapi_aead_getdata_output(struct kcapi_handle *handle,
+			       uint8_t *encdata, uint32_t encdatalen, int enc,
+			       uint8_t **aad, uint32_t *aadlen,
+			       uint8_t **data, uint32_t *datalen,
+			       uint8_t **tag, uint32_t *taglen);
+
 /**
  * kcapi_aead_decrypt() - synchronously decrypt AEAD data (one shot)
  *
@@ -607,17 +666,17 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  *
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
- * To catch authentication errors (i.e. integrity violations) during the
- * decryption operation, the errno of this call shall be checked for EBADMSG.
- * If this function returns < 0 and errno is set to EBADMSG, an authentication
- * error is detected.
+ * To catch authentication errors (i.e. integrity violations) during
+ * the decryption operation, the return value of this call should be
+ * checked. If this function returns -EBADMSG, an authentication error
+ * was detected.
  *
  * IMPORTANT NOTE: The kernel will only process
  * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
  * be handled by the kernel.
  *
  * @return number of bytes decrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 			   uint8_t *in, uint32_t inlen,
@@ -628,7 +687,8 @@ int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
  * kcapi_aead_decrypt_aio() - asynchronously decrypt AEAD data (one shot)
  *
  * @handle: [in] cipher handle
- * @iov: [in/out] array of scatter-gather list with input / output buffers
+ * @iniov: [in] array of scatter-gather list with input buffers
+ * @outiov: [out] array of scatter-gather list with output buffers
  * @iovlen: [in] number of IOVECs in array
  * @iv: [in] IV to be used for cipher operation
  * @access: [in] kernel access type (KCAPI_ACCESS_HEURISTIC - use internal
@@ -648,20 +708,21 @@ int32_t kcapi_aead_decrypt(struct kcapi_handle *handle,
  *
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
- * To catch authentication errors (i.e. integrity violations) during the
- * decryption operation, the errno of this call shall be checked for EBADMSG.
- * If this function returns < 0 and errno is set to EBADMSG, an authentication
- * error is detected.
+ * To catch authentication errors (i.e. integrity violations) during
+ * the decryption operation, the return value of this call should be
+ * checked. If this function returns -EBADMSG, an authentication error
+ * was detected.
  *
  * IMPORTANT NOTE: The kernel will only process
  * sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES at one time. Longer input data cannot
  * be handled by the kernel.
  *
  * @return number of bytes encrypted upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
-int32_t kcapi_aead_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
-			       uint32_t iovlen, const uint8_t *iv, int access);
+int32_t kcapi_aead_decrypt_aio(struct kcapi_handle *handle, struct iovec *iniov,
+			       struct iovec *outiov, uint32_t iovlen,
+			       const uint8_t *iv, int access);
 
 /**
  * kcapi_aead_stream_init_enc() - start an encryption operation (stream)
@@ -698,7 +759,7 @@ int32_t kcapi_aead_decrypt_aio(struct kcapi_handle *handle, struct iovec *iov,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
 				   const uint8_t *iv,
@@ -738,7 +799,7 @@ int32_t kcapi_aead_stream_init_enc(struct kcapi_handle *handle,
  * The IV buffer must be exactly kcapi_cipher_ivsize() bytes in size.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
 				   const uint8_t *iv,
@@ -776,7 +837,7 @@ int32_t kcapi_aead_stream_init_dec(struct kcapi_handle *handle,
  * be handled by the kernel.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
 				 struct iovec *iov, uint32_t iovlen);
@@ -797,7 +858,7 @@ int32_t kcapi_aead_stream_update(struct kcapi_handle *handle,
  * is triggered. Typically, the tag value is provided with this call.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
 				      struct iovec *iov, uint32_t iovlen);
@@ -819,7 +880,7 @@ int32_t kcapi_aead_stream_update_last(struct kcapi_handle *handle,
  * into the buffer.
  *
  * @return number of bytes obtained from the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
 			     struct iovec *iov, uint32_t iovlen);
@@ -830,7 +891,7 @@ int32_t kcapi_aead_stream_op(struct kcapi_handle *handle,
  * @handle: [in] cipher handle
  *
  * @return > 0 specifying the IV size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_aead_ivsize(struct kcapi_handle *handle);
 
@@ -840,7 +901,7 @@ uint32_t kcapi_aead_ivsize(struct kcapi_handle *handle);
  * @handle: [in] cipher handle
  *
  * @return > 0 specifying the block size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_aead_blocksize(struct kcapi_handle *handle);
 
@@ -854,7 +915,7 @@ uint32_t kcapi_aead_blocksize(struct kcapi_handle *handle);
  * the AEAD cipher type.
  *
  * @return > 0 specifying the block size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_aead_authsize(struct kcapi_handle *handle);
 
@@ -866,10 +927,68 @@ uint32_t kcapi_aead_authsize(struct kcapi_handle *handle);
  * @assoclen: [in] size of associated data (AD)
  * @taglen: [in] size of authentication tag
  *
+ * DEPRECATED since 0.13.0
+ *
  * @return minimum size of output data length in bytes
  */
-uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
-			    uint32_t inlen, uint32_t assoclen, uint32_t taglen);
+uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle, uint32_t inlen,
+			      uint32_t assoclen, uint32_t taglen);
+
+/**
+ * kcapi_aead_inbuflen_enc() - return minimum encryption input buffer length
+ *
+ * @handle: [in] cipher handle
+ * @inlen: [in] size of plaintext
+ * @assoclen: [in] size of associated data (AD)
+ * @taglen: [in] size of authentication tag
+ *
+ * @return minimum size of input data length in bytes
+ */
+uint32_t kcapi_aead_inbuflen_enc(struct kcapi_handle *handle,
+				 uint32_t inlen, uint32_t assoclen,
+				 uint32_t taglen);
+
+/**
+ * kcapi_aead_inbuflen_dec() - return minimum decryption input buffer length
+ *
+ * @handle: [in] cipher handle
+ * @inlen: [in] size of ciphertext
+ * @assoclen: [in] size of associated data (AD)
+ * @taglen: [in] size of authentication tag
+ *
+ * @return minimum size of output data length in bytes
+ */
+uint32_t kcapi_aead_inbuflen_dec(struct kcapi_handle *handle,
+				 uint32_t inlen, uint32_t assoclen,
+				 uint32_t taglen);
+
+/**
+ * kcapi_aead_outbuflen_enc() - return minimum encryption output buffer length
+ *
+ * @handle: [in] cipher handle
+ * @inlen: [in] size of plaintext
+ * @assoclen: [in] size of associated data (AD)
+ * @taglen: [in] size of authentication tag
+ *
+ * @return minimum size of output data length in bytes
+ */
+uint32_t kcapi_aead_outbuflen_enc(struct kcapi_handle *handle,
+				  uint32_t inlen, uint32_t assoclen,
+				  uint32_t taglen);
+
+/**
+ * kcapi_aead_outbuflen_dec() - return minimum decryption output buffer length
+ *
+ * @handle: [in] cipher handle
+ * @inlen: [in] size of ciphertext
+ * @assoclen: [in] size of associated data (AD)
+ * @taglen: [in] size of authentication tag
+ *
+ * @return minimum size of output data length in bytes
+ */
+uint32_t kcapi_aead_outbuflen_dec(struct kcapi_handle *handle,
+				  uint32_t inlen, uint32_t assoclen,
+				  uint32_t taglen);
 
 /**
  * kcapi_aead_ccm_nonce_to_iv() - convert CCM nonce into IV
@@ -885,7 +1004,7 @@ uint32_t kcapi_aead_outbuflen(struct kcapi_handle *handle,
  * Caller must free iv.
  *
  * @return 0 upon success;
- *	    < 0 upon failure
+ *	   < 0 upon failure
  */
 int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
 			       uint8_t **iv, uint32_t *ivlen);
@@ -905,10 +1024,14 @@ int kcapi_aead_ccm_nonce_to_iv(const uint8_t *nonce, uint32_t noncelen,
  * This function provides the initialization of a (keyed) message digest handle
  * and establishes the connection to the kernel.
  *
- * @return 0 upon success; ENOENT - algorithm not available;
- *	    -EOPNOTSUPP - AF_ALG family not available;
- *	    -EINVAL - accept syscall failed;
- *	    -ENOMEM - cipher handle cannot be allocated
+ * On success, a pointer to kcapi_handle object is returned in *handle.
+ * Function kcapi_md_destroy should be called afterwards to free resources.
+ *
+ * @return 0 upon success;
+ *	   -ENOENT - algorithm not available;
+ *	   -EOPNOTSUPP - AF_ALG family not available;
+ *	   -EINVAL - accept syscall failed;
+ *	   -ENOMEM - cipher handle cannot be allocated
  */
 int kcapi_md_init(struct kcapi_handle **handle, const char *ciphername,
 		  uint32_t flags);
@@ -934,7 +1057,7 @@ void kcapi_md_destroy(struct kcapi_handle *handle);
  * as it is now maintained by the kernel.
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_md_setkey(struct kcapi_handle *handle,
 		    const uint8_t *key, uint32_t keylen);
@@ -947,7 +1070,7 @@ int kcapi_md_setkey(struct kcapi_handle *handle,
  * @len: [in] buffer length
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_md_update(struct kcapi_handle *handle,
 			const uint8_t *buffer, uint32_t len);
@@ -1012,7 +1135,7 @@ uint32_t kcapi_md_digestsize(struct kcapi_handle *handle);
  * @handle: [in] cipher handle
  *
  * @return > 0 specifying the block size;
- *	    0 on error
+ *	   0 on error
  */
 uint32_t kcapi_md_blocksize(struct kcapi_handle *handle);
 
@@ -1032,10 +1155,14 @@ uint32_t kcapi_md_blocksize(struct kcapi_handle *handle);
  * This function provides the initialization of a random number generator handle
  * and establishes the connection to the kernel.
  *
- * @return 0 upon success; ENOENT - algorithm not available;
- *	    -EOPNOTSUPP - AF_ALG family not available;
- *	    -EINVAL - accept syscall failed
- *	    -ENOMEM - cipher handle cannot be allocated
+ * On success, a pointer to kcapi_handle object is returned in *handle.
+ * Function kcapi_rng_destroy should be called afterwards to free resources.
+ *
+ * @return 0 upon success;
+ *	   -ENOENT - algorithm not available;
+ *	   -EOPNOTSUPP - AF_ALG family not available;
+ *	   -EINVAL - accept syscall failed
+ *	   -ENOMEM - cipher handle cannot be allocated
  */
 int kcapi_rng_init(struct kcapi_handle **handle, const char *ciphername,
 		   uint32_t flags);
@@ -1063,7 +1190,7 @@ void kcapi_rng_destroy(struct kcapi_handle *handle);
  * in case of the initial seeding or additional data for reseeding.
  *
  * @return 0 upon success;
- * 	    < 0 upon error
+ * 	   a negative errno-style error code if an error occurred
  */
 int kcapi_rng_seed(struct kcapi_handle *handle, uint8_t *seed,
 		   uint32_t seedlen);
@@ -1076,7 +1203,7 @@ int kcapi_rng_seed(struct kcapi_handle *handle, uint8_t *seed,
  * @len: [in] buffer length
  *
  * @return size of random number generated upon success;
- *	    -EIO - data cannot be obtained
+ *	   -EIO - data cannot be obtained
  */
 int32_t kcapi_rng_generate(struct kcapi_handle *handle,
 			   uint8_t *buffer, uint32_t len);
@@ -1145,7 +1272,7 @@ uint32_t kcapi_version(void);
  * success. The consumer must free the memory after use.
  *
  * @return 0 for success;
- *	    < 0 for any errors
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_pad_iv(struct kcapi_handle *handle,
 		 const uint8_t *iv, uint32_t ivlen,
@@ -1180,10 +1307,15 @@ void kcapi_memset_secure(void *s, int c, uint32_t n);
  * This function provides the initialization of an asymmetric cipher handle and
  * establishes the connection to the kernel.
  *
- * @return 0 upon success; ENOENT - algorithm not available;
- *	    -EOPNOTSUPP - AF_ALG family not available;
- *	    -EINVAL - accept syscall failed
- *	    -ENOMEM - cipher handle cannot be allocated
+ * On success, a pointer to kcapi_handle object is returned in *handle.
+ * Function kcapi_akcipher_destroy should be called afterwards to free
+ * resources.
+ *
+ * @return 0 upon success;
+ *	   -ENOENT - algorithm not available;
+ *	   -EOPNOTSUPP - AF_ALG family not available;
+ *	   -EINVAL - accept syscall failed
+ *	   -ENOMEM - cipher handle cannot be allocated
  */
 int kcapi_akcipher_init(struct kcapi_handle **handle, const char *ciphername,
 			uint32_t flags);
@@ -1222,7 +1354,7 @@ void kcapi_akcipher_destroy(struct kcapi_handle *handle);
  * as it is now maintained by the kernel.
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_akcipher_setkey(struct kcapi_handle *handle,
 			  const uint8_t *key, uint32_t keylen);
@@ -1247,7 +1379,7 @@ int kcapi_akcipher_setkey(struct kcapi_handle *handle,
  * as it is now maintained by the kernel.
  *
  * @return 0 upon success;
- *	    < 0 in case of error
+ *	   a negative errno-style error code if an error occurred
  */
 int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
 			     const uint8_t *key, uint32_t keylen);
@@ -1276,7 +1408,7 @@ int kcapi_akcipher_setpubkey(struct kcapi_handle *handle,
  * output buffer must be at least as large as the modululs of the uses key.
  *
  * @return number of bytes returned by the encryption operation upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
@@ -1306,7 +1438,7 @@ int32_t kcapi_akcipher_encrypt(struct kcapi_handle *handle,
  * output buffer must be at least as large as the modululs of the uses key.
  *
  * @return number of bytes returned by the decryption operation upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
 			       const uint8_t *in, uint32_t inlen,
@@ -1336,7 +1468,7 @@ int32_t kcapi_akcipher_decrypt(struct kcapi_handle *handle,
  * output buffer must be at least as large as the modululs of the uses key.
  *
  * @return number of bytes returned by the signature gen operation upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
 			    const uint8_t *in, uint32_t inlen,
@@ -1365,12 +1497,12 @@ int32_t kcapi_akcipher_sign(struct kcapi_handle *handle,
  * If the output size is insufficiently large, -EINVAL is returned. The
  * output buffer must be at least as large as the modululs of the uses key.
  *
- * To catch signature verification errors, the errno of this call shall be
- * checked for EBADMSG. If this function returns < 0 and errno is set to
- * EBADMSG, the verification of the signature failed.
+ * To catch signature verification errors, the return value of this
+ * call should be checked. If this function returns -EBADMSG, the
+ * verification of the signature failed.
  *
  * @return number of bytes returned by the signature ver operation upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
 			      const uint8_t *in, uint32_t inlen,
@@ -1400,7 +1532,7 @@ int32_t kcapi_akcipher_verify(struct kcapi_handle *handle,
  * application.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
@@ -1428,7 +1560,7 @@ int32_t kcapi_akcipher_stream_init_enc(struct kcapi_handle *handle,
  * application.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
@@ -1457,7 +1589,7 @@ int32_t kcapi_akcipher_stream_init_dec(struct kcapi_handle *handle,
  * application.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
@@ -1487,7 +1619,7 @@ int32_t kcapi_akcipher_stream_init_sgn(struct kcapi_handle *handle,
  * application.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
 				       struct iovec *iov, uint32_t iovlen);
@@ -1513,7 +1645,7 @@ int32_t kcapi_akcipher_stream_init_vfy(struct kcapi_handle *handle,
  * processed by the cipher operation.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
 				     struct iovec *iov, uint32_t iovlen);
@@ -1537,7 +1669,7 @@ int32_t kcapi_akcipher_stream_update(struct kcapi_handle *handle,
  * notifies the kernel that no further data is to be expected.
  *
  * @return number of bytes sent to the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
 					  struct iovec *iov, uint32_t iovlen);
@@ -1563,7 +1695,7 @@ int32_t kcapi_akcipher_stream_update_last(struct kcapi_handle *handle,
  * into the buffer.
  *
  * @return number of bytes obtained from the kernel upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_akcipher_stream_op(struct kcapi_handle *handle,
 			         struct iovec *iov, uint32_t iovlen);
@@ -1598,7 +1730,7 @@ int32_t kcapi_akcipher_stream_op(struct kcapi_handle *handle,
  * may also be NULL if the caller wishes not to provide anything.
  *
  * @return 0 upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_kdf_dpi(struct kcapi_handle *handle,
 		      const uint8_t *src, uint32_t slen,
@@ -1628,7 +1760,7 @@ int32_t kcapi_kdf_dpi(struct kcapi_handle *handle,
  * hash) of the PRF.
  *
  * @return 0 upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_kdf_fb(struct kcapi_handle *handle,
 		     const uint8_t *src, uint32_t slen,
@@ -1661,7 +1793,7 @@ int32_t kcapi_kdf_fb(struct kcapi_handle *handle,
  * arbitrary string (see SP800-56A section 5.8.1.2).
  *
  * @return 0 upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_kdf_ctr(struct kcapi_handle *handle,
 		      const uint8_t *src, uint32_t slen,
@@ -1682,7 +1814,7 @@ int32_t kcapi_kdf_ctr(struct kcapi_handle *handle,
  * This function is an implementation of the PBKDF as defined in SP800-132.
  *
  * @return 0 upon success;
- *	    < 0 in case of error with errno set
+ *	   a negative errno-style error code if an error occurred
  */
 int32_t kcapi_pbkdf(const char *hashname,
 		    const uint8_t *pw, uint32_t pwlen,
